@@ -1,6 +1,6 @@
-use alloc::borrow::Cow;
-use alloc::fmt;
-use alloc::vec::Vec;
+use crate::alloc::borrow::Cow;
+use crate::alloc::fmt;
+use crate::alloc::vec::Vec;
 use std::iter;
 use std::slice;
 
@@ -14,9 +14,9 @@ use goblin::{elf, strtab};
 use scroll::ctx::TryFromCtx;
 use scroll::{self, Pread};
 
-use {
-    Machine, Object, ObjectSection, ObjectSegment, Relocation, RelocationKind, SectionKind, Symbol,
-    SymbolKind, SymbolMap,
+use crate::{
+    Machine, Object, ObjectSection, ObjectSegment, Relocation, RelocationKind, SectionIndex,
+    SectionKind, Symbol, SymbolIndex, SymbolKind, SymbolMap,
 };
 
 /// An ELF object file.
@@ -63,7 +63,7 @@ where
     'data: 'file,
 {
     file: &'file ElfFile<'data>,
-    index: usize,
+    index: SectionIndex,
     section: &'file elf::SectionHeader,
 }
 
@@ -73,8 +73,7 @@ where
     'data: 'file,
 {
     strtab: &'file strtab::Strtab<'data>,
-    symbols: elf::sym::SymIterator<'data>,
-    section_kinds: Vec<SectionKind>,
+    symbols: iter::Enumerate<elf::sym::SymIterator<'data>>,
 }
 
 /// An iterator over the relocations in an `ElfSection`.
@@ -83,7 +82,7 @@ where
     'data: 'file,
 {
     /// The index of the section that the relocations apply to.
-    section_index: u32,
+    section_index: SectionIndex,
     file: &'file ElfFile<'data>,
     sections: slice::Iter<'file, (elf::ShdrIdx, elf::RelocSection<'data>)>,
     relocations: Option<elf::reloc::RelocIterator<'data>>,
@@ -103,6 +102,12 @@ impl<'data> ElfFile<'data> {
         Ok(ElfFile { elf, data })
     }
 
+    /// True for 64-bit files.
+    #[inline]
+    pub fn is_64(&self) -> bool {
+        self.elf.is_64
+    }
+
     fn raw_section_by_name<'file>(
         &'file self,
         section_name: &str,
@@ -112,7 +117,7 @@ impl<'data> ElfFile<'data> {
                 if name == section_name {
                     return Some(ElfSection {
                         file: self,
-                        index,
+                        index: SectionIndex(index),
                         section,
                     });
                 }
@@ -135,7 +140,7 @@ impl<'data> ElfFile<'data> {
     #[cfg(not(feature = "compression"))]
     fn zdebug_section_by_name<'file>(
         &'file self,
-        section_name: &str,
+        _section_name: &str,
     ) -> Option<ElfSection<'data, 'file>> {
         None
     }
@@ -157,6 +162,7 @@ where
             elf::header::EM_AARCH64 => Machine::Arm64,
             elf::header::EM_386 => Machine::X86,
             elf::header::EM_X86_64 => Machine::X86_64,
+            elf::header::EM_MIPS => Machine::Mips,
             _ => Machine::Other,
         }
     }
@@ -173,6 +179,17 @@ where
             .or_else(|| self.zdebug_section_by_name(section_name))
     }
 
+    fn section_by_index(&'file self, index: SectionIndex) -> Option<ElfSection<'data, 'file>> {
+        self.elf
+            .section_headers
+            .get(index.0)
+            .map(|section| ElfSection {
+                file: self,
+                index,
+                section,
+            })
+    }
+
     fn sections(&'file self) -> ElfSectionIterator<'data, 'file> {
         ElfSectionIterator {
             file: self,
@@ -180,32 +197,33 @@ where
         }
     }
 
-    fn symbol_by_index(&self, index: u64) -> Option<Symbol<'data>> {
-        // TODO: determine section_kind too
+    fn symbol_by_index(&self, index: SymbolIndex) -> Option<Symbol<'data>> {
         self.elf
             .syms
-            .get(index as usize)
-            .map(|symbol| parse_symbol(&symbol, &self.elf.strtab, &[]))
+            .get(index.0)
+            .map(|symbol| parse_symbol(index.0, &symbol, &self.elf.strtab))
     }
 
     fn symbols(&'file self) -> ElfSymbolIterator<'data, 'file> {
         ElfSymbolIterator {
             strtab: &self.elf.strtab,
-            symbols: self.elf.syms.iter(),
-            section_kinds: self.sections().map(|x| x.kind()).collect(),
+            symbols: self.elf.syms.iter().enumerate(),
         }
     }
 
     fn dynamic_symbols(&'file self) -> ElfSymbolIterator<'data, 'file> {
         ElfSymbolIterator {
             strtab: &self.elf.dynstrtab,
-            symbols: self.elf.dynsyms.iter(),
-            section_kinds: self.sections().map(|x| x.kind()).collect(),
+            symbols: self.elf.dynsyms.iter().enumerate(),
         }
     }
 
     fn symbol_map(&self) -> SymbolMap<'data> {
-        let mut symbols: Vec<_> = self.symbols().filter(SymbolMap::filter).collect();
+        let mut symbols: Vec<_> = self
+            .symbols()
+            .map(|(_, s)| s)
+            .filter(SymbolMap::filter)
+            .collect();
         symbols.sort_by_key(|x| x.address);
         SymbolMap { symbols }
     }
@@ -299,10 +317,19 @@ impl<'data, 'file> ObjectSegment<'data> for ElfSegment<'data, 'file> {
         self.segment.p_memsz
     }
 
+    #[inline]
+    fn align(&self) -> u64 {
+        self.segment.p_align
+    }
+
     fn data(&self) -> &'data [u8] {
         &self.file.data[self.segment.p_offset as usize..][..self.segment.p_filesz as usize]
     }
 
+    fn data_range(&self, address: u64, size: u64) -> Option<&'data [u8]> {
+        crate::data_range(self.data(), self.address(), address, size)
+    }
+
     #[inline]
     fn name(&self) -> Option<&str> {
         None
@@ -314,7 +341,7 @@ impl<'data, 'file> Iterator for ElfSectionIterator<'data, 'file> {
 
     fn next(&mut self) -> Option<Self::Item> {
         self.iter.next().map(|(index, section)| ElfSection {
-            index,
+            index: SectionIndex(index),
             file: self.file,
             section,
         })
@@ -332,7 +359,7 @@ impl<'data, 'file> ElfSection<'data, 'file> {
 
     #[cfg(feature = "compression")]
     fn maybe_decompress_data(&self) -> Option<Cow<'data, [u8]>> {
-        if (self.section.sh_flags & elf::section_header::SHF_COMPRESSED as u64) == 0 {
+        if (self.section.sh_flags & u64::from(elf::section_header::SHF_COMPRESSED)) == 0 {
             return None;
         }
 
@@ -357,8 +384,9 @@ impl<'data, 'file> ElfSection<'data, 'file> {
 
         let mut decompressed = Vec::with_capacity(uncompressed_size as usize);
         let mut decompress = Decompress::new(true);
-        if let Err(_) =
-            decompress.decompress_vec(compressed_data, &mut decompressed, FlushDecompress::Finish)
+        if decompress
+            .decompress_vec(compressed_data, &mut decompressed, FlushDecompress::Finish)
+            .is_err()
         {
             return None;
         }
@@ -385,8 +413,9 @@ impl<'data, 'file> ElfSection<'data, 'file> {
         let uncompressed_size: u32 = data.pread_with(8, scroll::BE).unwrap();
         let mut decompressed = Vec::with_capacity(uncompressed_size as usize);
         let mut decompress = Decompress::new(true);
-        if let Err(_) =
-            decompress.decompress_vec(&data[12..], &mut decompressed, FlushDecompress::Finish)
+        if decompress
+            .decompress_vec(&data[12..], &mut decompressed, FlushDecompress::Finish)
+            .is_err()
         {
             return None;
         }
@@ -397,6 +426,11 @@ impl<'data, 'file> ElfSection<'data, 'file> {
 impl<'data, 'file> ObjectSection<'data> for ElfSection<'data, 'file> {
     type RelocationIterator = ElfRelocationIterator<'data, 'file>;
 
+    #[inline]
+    fn index(&self) -> SectionIndex {
+        self.index
+    }
+
     #[inline]
     fn address(&self) -> u64 {
         self.section.sh_addr
@@ -407,11 +441,20 @@ impl<'data, 'file> ObjectSection<'data> for ElfSection<'data, 'file> {
         self.section.sh_size
     }
 
+    #[inline]
+    fn align(&self) -> u64 {
+        self.section.sh_addralign
+    }
+
     #[inline]
     fn data(&self) -> Cow<'data, [u8]> {
         Cow::from(self.raw_data())
     }
 
+    fn data_range(&self, address: u64, size: u64) -> Option<&'data [u8]> {
+        crate::data_range(self.raw_data(), self.address(), address, size)
+    }
+
     #[cfg(feature = "compression")]
     fn uncompressed_data(&self) -> Cow<'data, [u8]> {
         self.maybe_decompress_data()
@@ -441,25 +484,52 @@ impl<'data, 'file> ObjectSection<'data> for ElfSection<'data, 'file> {
     fn kind(&self) -> SectionKind {
         match self.section.sh_type {
             elf::section_header::SHT_PROGBITS => {
-                if self.section.sh_flags & u64::from(elf::section_header::SHF_ALLOC) == 0 {
-                    SectionKind::Unknown
-                } else if self.section.sh_flags & u64::from(elf::section_header::SHF_EXECINSTR) != 0
-                {
-                    SectionKind::Text
-                } else if self.section.sh_flags & u64::from(elf::section_header::SHF_WRITE) != 0 {
-                    SectionKind::Data
+                if self.section.sh_flags & u64::from(elf::section_header::SHF_ALLOC) != 0 {
+                    if self.section.sh_flags & u64::from(elf::section_header::SHF_EXECINSTR) != 0 {
+                        SectionKind::Text
+                    } else if self.section.sh_flags & u64::from(elf::section_header::SHF_TLS) != 0 {
+                        SectionKind::Tls
+                    } else if self.section.sh_flags & u64::from(elf::section_header::SHF_WRITE) != 0
+                    {
+                        SectionKind::Data
+                    } else if self.section.sh_flags & u64::from(elf::section_header::SHF_STRINGS)
+                        != 0
+                    {
+                        SectionKind::ReadOnlyString
+                    } else {
+                        SectionKind::ReadOnlyData
+                    }
+                } else if self.section.sh_flags & u64::from(elf::section_header::SHF_STRINGS) != 0 {
+                    SectionKind::OtherString
                 } else {
-                    SectionKind::ReadOnlyData
+                    SectionKind::Other
                 }
             }
-            elf::section_header::SHT_NOBITS => SectionKind::UninitializedData,
-            _ => SectionKind::Unknown,
+            elf::section_header::SHT_NOBITS => {
+                if self.section.sh_flags & u64::from(elf::section_header::SHF_TLS) != 0 {
+                    SectionKind::UninitializedTls
+                } else {
+                    SectionKind::UninitializedData
+                }
+            }
+            elf::section_header::SHT_NULL
+            | elf::section_header::SHT_SYMTAB
+            | elf::section_header::SHT_STRTAB
+            | elf::section_header::SHT_RELA
+            | elf::section_header::SHT_HASH
+            | elf::section_header::SHT_DYNAMIC
+            | elf::section_header::SHT_REL
+            | elf::section_header::SHT_DYNSYM => SectionKind::Metadata,
+            _ => {
+                // TODO: maybe add more specialised kinds based on sh_type (e.g. Unwind)
+                SectionKind::Unknown
+            }
         }
     }
 
     fn relocations(&self) -> ElfRelocationIterator<'data, 'file> {
         ElfRelocationIterator {
-            section_index: self.index as u32,
+            section_index: self.index,
             file: self.file,
             sections: self.file.elf.shdr_relocs.iter(),
             relocations: None,
@@ -468,28 +538,32 @@ impl<'data, 'file> ObjectSection<'data> for ElfSection<'data, 'file> {
 }
 
 impl<'data, 'file> fmt::Debug for ElfSymbolIterator<'data, 'file> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("ElfSymbolIterator").finish()
     }
 }
 
 impl<'data, 'file> Iterator for ElfSymbolIterator<'data, 'file> {
-    type Item = Symbol<'data>;
+    type Item = (SymbolIndex, Symbol<'data>);
 
     fn next(&mut self) -> Option<Self::Item> {
-        self.symbols
-            .next()
-            .map(|symbol| parse_symbol(&symbol, self.strtab, &self.section_kinds))
+        self.symbols.next().map(|(index, symbol)| {
+            (
+                SymbolIndex(index),
+                parse_symbol(index, &symbol, self.strtab),
+            )
+        })
     }
 }
 
 fn parse_symbol<'data>(
+    index: usize,
     symbol: &elf::sym::Sym,
     strtab: &strtab::Strtab<'data>,
-    section_kinds: &[SectionKind],
 ) -> Symbol<'data> {
     let name = strtab.get(symbol.st_name).and_then(Result::ok);
     let kind = match elf::sym::st_type(symbol.st_info) {
+        elf::sym::STT_NOTYPE if index == 0 => SymbolKind::Null,
         elf::sym::STT_OBJECT => SymbolKind::Data,
         elf::sym::STT_FUNC => SymbolKind::Text,
         elf::sym::STT_SECTION => SymbolKind::Section,
@@ -498,17 +572,20 @@ fn parse_symbol<'data>(
         elf::sym::STT_TLS => SymbolKind::Tls,
         _ => SymbolKind::Unknown,
     };
-    let section_kind = if symbol.st_shndx == elf::section_header::SHN_UNDEF as usize {
-        None
-    } else {
-        Some(section_kinds.get(symbol.st_shndx).cloned().unwrap_or(SectionKind::Unknown))
-    };
+    let undefined = symbol.st_shndx == elf::section_header::SHN_UNDEF as usize;
+    let section_index =
+        if undefined || symbol.st_shndx >= elf::section_header::SHN_LORESERVE as usize {
+            None
+        } else {
+            Some(SectionIndex(symbol.st_shndx))
+        };
     Symbol {
         name,
         address: symbol.st_value,
         size: symbol.st_size,
         kind,
-        section_kind,
+        section_index,
+        undefined,
         global: elf::sym::st_bind(symbol.st_info) != elf::sym::STB_LOCAL,
     }
 }
@@ -519,34 +596,50 @@ impl<'data, 'file> Iterator for ElfRelocationIterator<'data, 'file> {
     fn next(&mut self) -> Option<Self::Item> {
         loop {
             if let Some(ref mut relocations) = self.relocations {
-                while let Some(reloc) = relocations.next() {
-                    let kind = match self.file.elf.header.e_machine {
+                if let Some(reloc) = relocations.next() {
+                    let (kind, size) = match self.file.elf.header.e_machine {
                         elf::header::EM_ARM => match reloc.r_type {
-                            elf::reloc::R_ARM_ABS32 => RelocationKind::Direct32,
-                            _ => RelocationKind::Other(reloc.r_type),
+                            elf::reloc::R_ARM_ABS32 => (RelocationKind::Absolute, 32),
+                            _ => (RelocationKind::Other(reloc.r_type), 0),
                         },
                         elf::header::EM_AARCH64 => match reloc.r_type {
-                            elf::reloc::R_AARCH64_ABS64 => RelocationKind::Direct64,
-                            elf::reloc::R_AARCH64_ABS32 => RelocationKind::Direct32,
-                            _ => RelocationKind::Other(reloc.r_type),
+                            elf::reloc::R_AARCH64_ABS64 => (RelocationKind::Absolute, 64),
+                            elf::reloc::R_AARCH64_ABS32 => (RelocationKind::Absolute, 32),
+                            elf::reloc::R_AARCH64_ABS16 => (RelocationKind::Absolute, 16),
+                            elf::reloc::R_AARCH64_PREL64 => (RelocationKind::Relative, 64),
+                            elf::reloc::R_AARCH64_PREL32 => (RelocationKind::Relative, 32),
+                            elf::reloc::R_AARCH64_PREL16 => (RelocationKind::Relative, 16),
+                            _ => (RelocationKind::Other(reloc.r_type), 0),
                         },
                         elf::header::EM_386 => match reloc.r_type {
-                            elf::reloc::R_386_32 => RelocationKind::Direct32,
-                            _ => RelocationKind::Other(reloc.r_type),
+                            elf::reloc::R_386_32 => (RelocationKind::Absolute, 32),
+                            elf::reloc::R_386_PC32 => (RelocationKind::Relative, 32),
+                            elf::reloc::R_386_GOT32 => (RelocationKind::GotRelative, 32),
+                            elf::reloc::R_386_PLT32 => (RelocationKind::PltRelative, 32),
+                            _ => (RelocationKind::Other(reloc.r_type), 0),
                         },
                         elf::header::EM_X86_64 => match reloc.r_type {
-                            elf::reloc::R_X86_64_64 => RelocationKind::Direct64,
-                            elf::reloc::R_X86_64_32 => RelocationKind::Direct32,
-                            elf::reloc::R_X86_64_32S => RelocationKind::DirectSigned32,
-                            _ => RelocationKind::Other(reloc.r_type),
+                            elf::reloc::R_X86_64_64 => (RelocationKind::Absolute, 64),
+                            elf::reloc::R_X86_64_PC32 => (RelocationKind::Relative, 32),
+                            elf::reloc::R_X86_64_GOT32 => (RelocationKind::GotOffset, 32),
+                            elf::reloc::R_X86_64_PLT32 => (RelocationKind::PltRelative, 32),
+                            elf::reloc::R_X86_64_GOTPCREL => (RelocationKind::GotRelative, 32),
+                            elf::reloc::R_X86_64_32 => (RelocationKind::Absolute, 32),
+                            elf::reloc::R_X86_64_32S => (RelocationKind::AbsoluteSigned, 32),
+                            elf::reloc::R_X86_64_16 => (RelocationKind::Absolute, 16),
+                            elf::reloc::R_X86_64_PC16 => (RelocationKind::Relative, 16),
+                            elf::reloc::R_X86_64_8 => (RelocationKind::Absolute, 8),
+                            elf::reloc::R_X86_64_PC8 => (RelocationKind::Relative, 8),
+                            _ => (RelocationKind::Other(reloc.r_type), 0),
                         },
-                        _ => RelocationKind::Other(reloc.r_type),
+                        _ => (RelocationKind::Other(reloc.r_type), 0),
                     };
                     return Some((
                         reloc.r_offset,
                         Relocation {
                             kind,
-                            symbol: reloc.r_sym as u64,
+                            size,
+                            symbol: SymbolIndex(reloc.r_sym as usize),
                             addend: reloc.r_addend.unwrap_or(0),
                             implicit_addend: reloc.r_addend.is_none(),
                         },
@@ -557,7 +650,7 @@ impl<'data, 'file> Iterator for ElfRelocationIterator<'data, 'file> {
                 None => return None,
                 Some((index, relocs)) => {
                     let section = &self.file.elf.section_headers[*index];
-                    if section.sh_info == self.section_index {
+                    if section.sh_info as usize == self.section_index.0 {
                         self.relocations = Some(relocs.into_iter());
                     }
                     // TODO: do we need to return section.sh_link?
@@ -568,7 +661,7 @@ impl<'data, 'file> Iterator for ElfRelocationIterator<'data, 'file> {
 }
 
 impl<'data, 'file> fmt::Debug for ElfRelocationIterator<'data, 'file> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("ElfRelocationIterator").finish()
     }
 }
