@@ -22,16 +22,6 @@ extern crate alloc;
 #[cfg(not(feature = "std"))]
 extern crate core as std;
 
-#[cfg(feature = "compression")]
-extern crate flate2;
-
-extern crate goblin;
-extern crate scroll;
-extern crate uuid;
-
-#[cfg(feature = "wasm")]
-extern crate parity_wasm;
-
 #[cfg(feature = "std")]
 mod alloc {
     pub use std::borrow;
@@ -39,26 +29,26 @@ mod alloc {
     pub use std::vec;
 }
 
-use alloc::borrow::Cow;
-use alloc::fmt;
-use alloc::vec::Vec;
+use crate::alloc::borrow::Cow;
+use crate::alloc::fmt;
+use crate::alloc::vec::Vec;
 
 mod elf;
-pub use elf::*;
+pub use crate::elf::*;
 
 mod macho;
-pub use macho::*;
+pub use crate::macho::*;
 
 mod pe;
-pub use pe::*;
+pub use crate::pe::*;
 
 mod traits;
-pub use traits::*;
+pub use crate::traits::*;
 
 #[cfg(feature = "wasm")]
 mod wasm;
 #[cfg(feature = "wasm")]
-pub use wasm::*;
+pub use crate::wasm::*;
 
 pub use uuid::Uuid;
 
@@ -78,12 +68,33 @@ pub type NativeFile<'data> = PeFile<'data>;
 #[cfg(all(feature = "wasm", target_arch = "wasm32"))]
 pub type NativeFile<'data> = WasmFile<'data>;
 
+/// The object file format.
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub enum Format {
+    /// 32-bit ELF
+    Elf32,
+    /// 64-bit ELF
+    Elf64,
+    /// 32-bit Mach-O
+    MachO32,
+    /// 64-bit Mach-O
+    MachO64,
+    /// 32-bit PE
+    Pe32,
+    /// 64-bit PE
+    Pe64,
+    /// WebAssembly
+    #[cfg(feature = "wasm")]
+    Wasm,
+}
+
 /// An object file.
 #[derive(Debug)]
 pub struct File<'data> {
     inner: FileInternal<'data>,
 }
 
+#[allow(clippy::large_enum_variant)]
 #[derive(Debug)]
 enum FileInternal<'data> {
     Elf(ElfFile<'data>),
@@ -107,6 +118,8 @@ pub enum Machine {
     /// x86-64
     #[allow(non_camel_case_types)]
     X86_64,
+    /// MIPS
+    Mips,
 }
 
 /// An iterator over the segments of a `File`.
@@ -172,6 +185,10 @@ where
     Wasm(WasmSectionIterator<'file>),
 }
 
+/// The index used to identify a section of a file.
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+pub struct SectionIndex(pub usize);
+
 /// A Section of a File
 pub struct Section<'data, 'file>
 where
@@ -197,15 +214,45 @@ pub enum SectionKind {
     /// The section kind is unknown.
     Unknown,
     /// An executable code section.
+    ///
+    /// Example ELF sections: `.text`
     Text,
     /// A data section.
+    ///
+    /// Example ELF sections: `.data`
     Data,
     /// A read only data section.
+    ///
+    /// Example ELF sections: `.rodata`
     ReadOnlyData,
+    /// A loadable string section.
+    ///
+    /// Example ELF sections: `.rodata.str`
+    ReadOnlyString,
     /// An uninitialized data section.
+    ///
+    /// Example ELF sections: `.bss`
     UninitializedData,
-    /// Some other type of text or data section.
+    /// A TLS data section.
+    ///
+    /// Example ELF sections: `.tdata`
+    Tls,
+    /// An uninitialized TLS data section.
+    ///
+    /// Example ELF sections: `.tbss`
+    UninitializedTls,
+    /// A non-loadable string section.
+    ///
+    /// Example ELF sections: `.comment`, `.debug_str`
+    OtherString,
+    /// Some other non-loadable section.
+    ///
+    /// Example ELF sections: `.debug_info`
     Other,
+    /// Metadata such as symbols or relocations.
+    ///
+    /// Example ELF sections: `.symtab`, `.strtab`
+    Metadata,
 }
 
 /// An iterator over symbol table entries.
@@ -229,11 +276,16 @@ where
     Wasm(WasmSymbolIterator<'file>),
 }
 
+/// The index used to identify a symbol of a file.
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+pub struct SymbolIndex(pub usize);
+
 /// A symbol table entry.
 #[derive(Debug)]
 pub struct Symbol<'data> {
     kind: SymbolKind,
-    section_kind: Option<SectionKind>,
+    section_index: Option<SectionIndex>,
+    undefined: bool,
     global: bool,
     name: Option<&'data str>,
     address: u64,
@@ -245,6 +297,8 @@ pub struct Symbol<'data> {
 pub enum SymbolKind {
     /// The symbol kind is unknown.
     Unknown,
+    /// The symbol is a null placeholder.
+    Null,
     /// The symbol is for executable code.
     Text,
     /// The symbol is for a data object.
@@ -269,20 +323,37 @@ pub struct SymbolMap<'data> {
 #[derive(Debug)]
 pub struct Relocation {
     kind: RelocationKind,
-    symbol: u64,
+    size: u8,
+    symbol: SymbolIndex,
     addend: i64,
     implicit_addend: bool,
 }
 
 /// The kind of a relocation.
+///
+/// The relocation descriptions use the following definitions. Note that
+/// these definitions probably don't match any ELF ABI.
+///
+/// * A - The value of the addend.
+/// * G - The address of the symbol's entry within the global offset table.
+/// * GOT - The address of the global offset table.
+/// * L - The address of the symbol's entry within the procedure linkage table.
+/// * P - The address of the place of the relocation.
+/// * S - The address of the symbol.
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum RelocationKind {
-    /// u32, symbol + addend
-    Direct32,
-    /// i32, symbol + addend
-    DirectSigned32,
-    /// u64, symbol + addend
-    Direct64,
+    /// S + A
+    Absolute,
+    /// S + A
+    AbsoluteSigned,
+    /// S + A - P
+    Relative,
+    /// G + A - GOT
+    GotOffset,
+    /// G + A - P
+    GotRelative,
+    /// L + A - P
+    PltRelative,
     /// Some other kind of relocation. The value is dependent on file format and machine.
     Other(u32),
 }
@@ -375,7 +446,7 @@ macro_rules! next_inner {
 }
 
 #[cfg(feature = "wasm")]
-fn parse_wasm(data: &[u8]) -> Result<Option<File>, &'static str> {
+fn parse_wasm(data: &[u8]) -> Result<Option<File<'_>>, &'static str> {
     const WASM_MAGIC: &[u8] = &[0x00, 0x61, 0x73, 0x6D];
 
     if &data[..4] == WASM_MAGIC {
@@ -412,6 +483,35 @@ impl<'data> File<'data> {
         };
         Ok(File { inner })
     }
+
+    /// Return the file format.
+    pub fn format(&self) -> Format {
+        match self.inner {
+            FileInternal::Elf(ref inner) => {
+                if inner.is_64() {
+                    Format::Elf64
+                } else {
+                    Format::Elf32
+                }
+            }
+            FileInternal::MachO(ref inner) => {
+                if inner.is_64() {
+                    Format::MachO64
+                } else {
+                    Format::MachO32
+                }
+            }
+            FileInternal::Pe(ref inner) => {
+                if inner.is_64() {
+                    Format::Pe64
+                } else {
+                    Format::Pe32
+                }
+            }
+            #[cfg(feature = "wasm")]
+            FileInternal::Wasm(_) => Format::Wasm,
+        }
+    }
 }
 
 impl<'data, 'file> Object<'data, 'file> for File<'data>
@@ -437,7 +537,14 @@ where
 
     fn section_by_name(&'file self, section_name: &str) -> Option<Section<'data, 'file>> {
         map_inner_option!(self.inner, FileInternal, SectionInternal, |x| x
-            .section_by_name(section_name)).map(|inner| Section { inner })
+            .section_by_name(section_name))
+        .map(|inner| Section { inner })
+    }
+
+    fn section_by_index(&'file self, index: SectionIndex) -> Option<Section<'data, 'file>> {
+        map_inner_option!(self.inner, FileInternal, SectionInternal, |x| x
+            .section_by_index(index))
+        .map(|inner| Section { inner })
     }
 
     fn section_data_by_name(&self, section_name: &str) -> Option<Cow<'data, [u8]>> {
@@ -452,7 +559,7 @@ where
         }
     }
 
-    fn symbol_by_index(&self, index: u64) -> Option<Symbol<'data>> {
+    fn symbol_by_index(&self, index: SymbolIndex) -> Option<Symbol<'data>> {
         with_inner!(self.inner, FileInternal, |x| x.symbol_by_index(index))
     }
 
@@ -512,7 +619,7 @@ impl<'data, 'file> Iterator for SegmentIterator<'data, 'file> {
 }
 
 impl<'data, 'file> fmt::Debug for Segment<'data, 'file> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         // It's painful to do much better than this
         f.debug_struct("Segment")
             .field("name", &self.name().unwrap_or("<unnamed>"))
@@ -531,10 +638,18 @@ impl<'data, 'file> ObjectSegment<'data> for Segment<'data, 'file> {
         with_inner!(self.inner, SegmentInternal, |x| x.size())
     }
 
+    fn align(&self) -> u64 {
+        with_inner!(self.inner, SegmentInternal, |x| x.align())
+    }
+
     fn data(&self) -> &'data [u8] {
         with_inner!(self.inner, SegmentInternal, |x| x.data())
     }
 
+    fn data_range(&self, address: u64, size: u64) -> Option<&'data [u8]> {
+        with_inner!(self.inner, SegmentInternal, |x| x.data_range(address, size))
+    }
+
     fn name(&self) -> Option<&str> {
         with_inner!(self.inner, SegmentInternal, |x| x.name())
     }
@@ -550,7 +665,7 @@ impl<'data, 'file> Iterator for SectionIterator<'data, 'file> {
 }
 
 impl<'data, 'file> fmt::Debug for Section<'data, 'file> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         // It's painful to do much better than this
         f.debug_struct("Section")
             .field("name", &self.name().unwrap_or("<invalid name>"))
@@ -564,6 +679,10 @@ impl<'data, 'file> fmt::Debug for Section<'data, 'file> {
 impl<'data, 'file> ObjectSection<'data> for Section<'data, 'file> {
     type RelocationIterator = RelocationIterator<'data, 'file>;
 
+    fn index(&self) -> SectionIndex {
+        with_inner!(self.inner, SectionInternal, |x| x.index())
+    }
+
     fn address(&self) -> u64 {
         with_inner!(self.inner, SectionInternal, |x| x.address())
     }
@@ -572,10 +691,18 @@ impl<'data, 'file> ObjectSection<'data> for Section<'data, 'file> {
         with_inner!(self.inner, SectionInternal, |x| x.size())
     }
 
+    fn align(&self) -> u64 {
+        with_inner!(self.inner, SectionInternal, |x| x.align())
+    }
+
     fn data(&self) -> Cow<'data, [u8]> {
         with_inner!(self.inner, SectionInternal, |x| x.data())
     }
 
+    fn data_range(&self, address: u64, size: u64) -> Option<&'data [u8]> {
+        with_inner!(self.inner, SectionInternal, |x| x.data_range(address, size))
+    }
+
     fn uncompressed_data(&self) -> Cow<'data, [u8]> {
         with_inner!(self.inner, SectionInternal, |x| x.uncompressed_data())
     }
@@ -605,7 +732,7 @@ impl<'data, 'file> ObjectSection<'data> for Section<'data, 'file> {
 }
 
 impl<'data, 'file> Iterator for SymbolIterator<'data, 'file> {
-    type Item = Symbol<'data>;
+    type Item = (SymbolIndex, Symbol<'data>);
 
     fn next(&mut self) -> Option<Self::Item> {
         with_inner_mut!(self.inner, SymbolIteratorInternal, |x| x.next())
@@ -619,16 +746,18 @@ impl<'data> Symbol<'data> {
         self.kind
     }
 
-    /// Returns the section kind for the symbol, or `None` if the symbol is undefined.
+    /// Returns the section index for the section containing this symbol.
+    ///
+    /// May return `None` if the section is unknown or the symbol is undefined.
     #[inline]
-    pub fn section_kind(&self) -> Option<SectionKind> {
-        self.section_kind
+    pub fn section_index(&self) -> Option<SectionIndex> {
+        self.section_index
     }
 
     /// Return true if the symbol is undefined.
     #[inline]
     pub fn is_undefined(&self) -> bool {
-        self.section_kind.is_none()
+        self.undefined
     }
 
     /// Return true if the symbol is global.
@@ -674,7 +803,8 @@ impl<'data> SymbolMap<'data> {
                 } else {
                     std::cmp::Ordering::Less
                 }
-            }).ok()
+            })
+            .ok()
             .and_then(|index| self.symbols.get(index))
     }
 
@@ -684,11 +814,15 @@ impl<'data> SymbolMap<'data> {
     }
 
     /// Return true for symbols that should be included in the map.
-    fn filter(symbol: &Symbol) -> bool {
+    fn filter(symbol: &Symbol<'_>) -> bool {
         match symbol.kind() {
             SymbolKind::Unknown | SymbolKind::Text | SymbolKind::Data => {}
-            SymbolKind::Section | SymbolKind::File | SymbolKind::Common | SymbolKind::Tls => {
-                return false
+            SymbolKind::Null
+            | SymbolKind::Section
+            | SymbolKind::File
+            | SymbolKind::Common
+            | SymbolKind::Tls => {
+                return false;
             }
         }
         !symbol.is_undefined() && symbol.size() > 0
@@ -710,9 +844,17 @@ impl Relocation {
         self.kind
     }
 
+    /// The size in bits of the place of the relocation.
+    ///
+    /// If 0, then the size is determined by the relocation kind.
+    #[inline]
+    pub fn size(&self) -> u8 {
+        self.size
+    }
+
     /// The index of the symbol within the symbol table, if applicable.
     #[inline]
-    pub fn symbol(&self) -> u64 {
+    pub fn symbol(&self) -> SymbolIndex {
         self.symbol
     }
 
@@ -732,3 +874,14 @@ impl Relocation {
         self.implicit_addend
     }
 }
+
+fn data_range(data: &[u8], data_address: u64, range_address: u64, size: u64) -> Option<&[u8]> {
+    if range_address >= data_address {
+        let start_offset = (range_address - data_address) as usize;
+        let end_offset = start_offset + size as usize;
+        if end_offset <= data.len() {
+            return Some(&data[start_offset..end_offset]);
+        }
+    }
+    None
+}
